# -*- coding: utf-8 -*-
from typing import Dict

SOLANA_PAY_PROTOCOL = "solana:"
SOLANA_ACTIONS_PROTOCOL = "solana-action:"
SOLANA_ACTIONS_PROTOCOL_PLURAL = "solana-actions:"
HTTPS_PROTOCOL = "https:"
MEMO_PROGRAM_ID = "MemoSq4gqABAXKb96qnH8TysNcWxMyWCqXgDLGmfcHr"
BLINKS_QUERY_PARAM = "action"

ACTIONS_CORS_HEADERS: Dict[str, str] = {
    "Access-Control-Allow-Origin": "*",
    "Access-Control-Allow-Methods": "GET,POST,PUT,OPTIONS",
    "Access-Control-Allow-Headers": "Content-Type, Authorization, Content-Encoding, Accept-Encoding",
    "Content-Type": "application/json",
}



# -*- coding: utf-8 -*-
from typing import Union
from urllib.parse import parse_qs, unquote, urlparse

from pydantic import BaseModel

from .constants import (
    BLINKS_QUERY_PARAM,
    HTTPS_PROTOCOL,
    SOLANA_ACTIONS_PROTOCOL,
    SOLANA_ACTIONS_PROTOCOL_PLURAL,
    SOLANA_PAY_PROTOCOL,
)


class ParseURLError(Exception):
    pass


class ActionRequestURLFields(BaseModel):
    link: str
    label: str | None = None
    message: str | None = None


class BlinkURLFields(BaseModel):
    blink: str
    action: ActionRequestURLFields


def parse_url(url: str) -> Union[ActionRequestURLFields, BlinkURLFields]:
    if len(url) > 2048:
        raise ParseURLError("length invalid")

    parsed_url = urlparse(url)

    if parsed_url.scheme in ("http", "https"):
        return parse_blink_url(parsed_url)

    if parsed_url.scheme not in (
        SOLANA_PAY_PROTOCOL[:-1],
        SOLANA_ACTIONS_PROTOCOL[:-1],
        SOLANA_ACTIONS_PROTOCOL_PLURAL[:-1],
    ):
        raise ParseURLError("protocol invalid")

    if not parsed_url.path:
        raise ParseURLError("pathname missing")

    if not any(char in parsed_url.path for char in [":", "%"]):
        raise ParseURLError("pathname invalid")

    return parse_action_request_url(parsed_url)


def parse_action_request_url(parsed_url) -> ActionRequestURLFields:
    path = unquote(parsed_url.path)
    link = urlparse(path)
    if link.scheme != HTTPS_PROTOCOL[:-1]:
        raise ParseURLError("link invalid")

    query_params = parse_qs(parsed_url.query)

    return ActionRequestURLFields(
        link=path,
        label=query_params.get("label", [None])[0],
        message=query_params.get("message", [None])[0],
    )


def parse_blink_url(parsed_url) -> BlinkURLFields:
    query_params = parse_qs(parsed_url.query)
    link = query_params.get(BLINKS_QUERY_PARAM, [None])[0]

    if not link:
        raise ParseURLError("invalid blink url")

    return BlinkURLFields(blink=(parsed_url.geturl()), action=parse_url(link))



# -*- coding: utf-8 -*-
import base64
from typing import Any, Dict, Optional, Union

import nacl.signing
import requests
from pydantic import BaseModel
from solana.rpc.async_api import AsyncClient
from solana.transaction import Transaction
from solders.pubkey import Pubkey


class FetchActionError(Exception):
    pass


class SerializeTransactionError(Exception):
    pass


class ActionPostRequest(BaseModel):
    account: str


class ActionPostResponse(BaseModel):
    transaction: str
    message: Optional[str] = None


class ActionPostResponseWithSerializedTransaction(ActionPostResponse):
    transaction: Transaction


async def fetch_transaction(
    connection: AsyncClient,
    link: str,
    fields: ActionPostRequest,
    options: Dict[str, Any] = {},
) -> ActionPostResponseWithSerializedTransaction:
    response = requests.post(
        str(link),
        json=fields.dict(),
        headers={
            "Accept": "application/json",
            "Content-Type": "application/json",
        },
        timeout=5,
    )
    json_data = response.json()

    if not json_data.get("transaction"):
        raise FetchActionError("missing transaction")
    if not isinstance(json_data["transaction"], str):
        raise FetchActionError("invalid transaction")

    transaction = await serialize_transaction(
        connection, fields.account, json_data["transaction"], options
    )

    return ActionPostResponseWithSerializedTransaction(
        **json_data, transaction=transaction
    )


async def serialize_transaction(
    connection: AsyncClient,
    account: Union[str, Pubkey],
    base64_transaction: str,
    options: Dict[str, Any] = {},
) -> Transaction:
    if isinstance(account, str):
        account = Pubkey(account)

    transaction = Transaction.deserialize(base64.b64decode(base64_transaction))
    signatures = transaction.signatures
    fee_payer = transaction.fee_payer()
    recent_blockhash = transaction.recent_blockhash

    if signatures:
        if not fee_payer:
            raise SerializeTransactionError("missing fee payer")
        if fee_payer != signatures[0].pubkey:
            raise SerializeTransactionError("invalid fee payer")
        if not recent_blockhash:
            raise SerializeTransactionError("missing recent blockhash")

        message = transaction.serialize_message()
        for sig in signatures:
            if sig.signature:
                if not nacl.signing.VerifyKey(bytes(sig.pubkey)).verify(
                    message, sig.signature
                ):
                    raise SerializeTransactionError("invalid signature")
            elif sig.pubkey == account:
                if len(signatures) == 1:
                    transaction.recent_blockhash = (
                        await connection.get_recent_blockhash(
                            commitment=options.get("commitment")
                        )
                    ).value.blockhash
            else:
                raise SerializeTransactionError("missing signature")
    else:
        transaction.fee_payer = account
        transaction.recent_blockhash = (
            await connection.get_recent_blockhash(commitment=options.get("commitment"))
        ).value.blockhash

    return transaction



# -*- coding: utf-8 -*-
from typing import Union
from urllib.parse import urlencode

from .constants import BLINKS_QUERY_PARAM, SOLANA_ACTIONS_PROTOCOL
from .types import (
    ActionRequestURLFields,
    BlinkURLFields,
    SupportedProtocols,
)


class EncodeURLError(Exception):
    pass


def encode_url(
    fields: Union[ActionRequestURLFields, BlinkURLFields],
    protocol: SupportedProtocols = SOLANA_ACTIONS_PROTOCOL,
) -> str:
    if isinstance(fields, BlinkURLFields):
        return encode_blink_url(fields, protocol)
    return encode_action_request_url(fields, protocol)


def encode_action_request_url(
    fields: ActionRequestURLFields,
    protocol: SupportedProtocols = SOLANA_ACTIONS_PROTOCOL,
) -> str:

    # TODO: do we maybe need a urllib.parse.quote around this?
    pathname = str(fields.link).rstrip("/")
    print(protocol)
    url = f"{protocol}{pathname}"

    params = {}
    if fields.label:
        params["label"] = fields.label
    if fields.message:
        params["message"] = fields.message

    if params:
        url += "?" + urlencode(params)

    return url


def encode_blink_url(
    fields: BlinkURLFields, protocol: SupportedProtocols = SOLANA_ACTIONS_PROTOCOL
) -> str:
    blink_url = str(fields.blink)
    action_url = encode_action_request_url(fields.action, protocol)

    separator = "&" if "?" in blink_url else "?"
    # TODO: do we maybe need a urllib.parse.quote around action_url?
    return f"{blink_url}{separator}{BLINKS_QUERY_PARAM}={(action_url)}"



# -*- coding: utf-8 -*-
from typing import List, Literal, Optional

from pydantic import BaseModel


class ActionsJson(BaseModel):
    rules: List["ActionRuleObject"]


class ActionRuleObject(BaseModel):
    pathPattern: str
    apiPath: str


class ActionRequestURLFields(BaseModel):
    link: str
    label: Optional[str] = None
    message: Optional[str] = None


class BlinkURLFields(BaseModel):
    blink: str
    action: ActionRequestURLFields


class ActionParameter(BaseModel):
    name: str
    label: Optional[str] = None
    required: Optional[bool] = False


class LinkedAction(BaseModel):
    href: str
    label: str
    parameters: Optional[List[ActionParameter]] = None


class ActionError(BaseModel):
    message: str


class ActionGetResponse(BaseModel):
    icon: str
    title: str
    description: str
    label: str
    disabled: Optional[bool] = None
    links: Optional[dict] = None
    error: Optional[ActionError] = None


class ActionPostRequest(BaseModel):
    account: str


class ActionPostResponse(BaseModel):
    transaction: str
    message: Optional[str] = None


# Pydantic cannot use constants loaded from .constants here, unfortunately
SupportedProtocols: Optional[
    Literal["solana-actions://", "solana-action://", "solana://"]
] = None
ActionsJson.update_forward_refs()



# -*- coding: utf-8 -*-
from typing import Any, Dict

from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Confirmed
from solders.pubkey import Pubkey


class FindReferenceError(Exception):
    pass


async def find_reference(
    connection: AsyncClient, reference: Pubkey, options: Dict[str, Any] = {}
):
    finality = options.pop("finality", Confirmed)

    signatures = await connection.get_signatures_for_address(
        reference,
        limit=options.get("limit", 1000),
        before=options.get("before"),
        until=options.get("until"),
        commitment=finality,
    )

    if not signatures:
        raise FindReferenceError("not found")

    oldest = signatures[-1]

    if len(signatures) < options.get("limit", 1000):
        return oldest

    try:
        return await find_reference(
            connection,
            reference,
            {
                "finality": finality,
                **options,
                "before": oldest.signature,
            },
        )
    except FindReferenceError:
        return oldest



# -*- coding: utf-8 -*-
from typing import List, Optional

from pydantic import BaseModel
from solders.keypair import Keypair
from solders.pubkey import Pubkey

from .action_identity import (
    create_action_identifier_instruction,
    get_action_identity_from_env,
)
from .constants import MEMO_PROGRAM_ID
from .types import ActionPostResponse, Reference


class CreatePostResponseError(Exception):
    pass


class CreateActionPostResponseArgs(BaseModel):
    fields: ActionPostResponse
    signers: Optional[List[Keypair]] = None
    action_identity: Optional[Keypair] = None
    reference: Optional[Reference] = None
    options: Optional[dict] = None


async def create_post_response(
    args: CreateActionPostResponseArgs,
) -> ActionPostResponse:
    transaction = args.fields.transaction

    if not transaction.recent_blockhash:
        transaction.recent_blockhash = "11111111111111111111111111111111"

    if not args.action_identity:
        try:
            args.action_identity = get_action_identity_from_env()
        except Exception:
            args.action_identity = None

    if len(transaction.instructions) <= 0:
        raise CreatePostResponseError("at least 1 instruction is required")

    if args.action_identity:
        instruction_data = create_action_identifier_instruction(
            args.action_identity, args.reference
        )
        transaction.add(instruction_data["instruction"])

        memo_id = Pubkey(MEMO_PROGRAM_ID)
        non_memo_index = next(
            (
                i
                for i, ix in enumerate(transaction.instructions)
                if ix.program_id != memo_id
            ),
            -1,
        )
        if non_memo_index == -1:
            raise CreatePostResponseError(
                "transaction requires at least 1 non-memo instruction"
            )

        transaction.instructions[non_memo_index].keys.extend(
            [
                {
                    "pubkey": args.action_identity.public_key,
                    "is_signer": False,
                    "is_writable": False,
                },
                {
                    "pubkey": instruction_data["reference"],
                    "is_signer": False,
                    "is_writable": False,
                },
            ]
        )

    if args.signers:
        for signer in args.signers:
            transaction.sign(signer)

    return ActionPostResponse(
        **args.fields.dict(), transaction=transaction.serialize().hex()
    )



# -*- coding: utf-8 -*-
from typing import List, Optional, Union

import base58
import nacl.signing
from pydantic import BaseModel
from solana.rpc.async_api import AsyncClient
from solana.transaction import TransactionInstruction
from solders.keypair import Keypair
from solders.pubkey import Pubkey

from .constants import MEMO_PROGRAM_ID, SOLANA_ACTIONS_PROTOCOL
from .find_reference import find_reference
from .types import Reference


class ActionsIdentitySchema(BaseModel):
    separator: str = ":"
    protocol: str = SOLANA_ACTIONS_PROTOCOL.replace(":", "")
    scheme: dict = {
        "protocol": 0,
        "identity": 1,
        "reference": 2,
        "signature": 3,
    }


ACTIONS_IDENTITY_SCHEMA = ActionsIdentitySchema()


class ActionIdentifierError(Exception):
    pass


def create_action_identifier_memo(identity: Keypair, reference: Reference) -> str:
    signature = nacl.signing.SigningKey(identity.secret_key()).sign(
        reference.to_bytes()
    )[:64]

    identifier = [""] * len(ACTIONS_IDENTITY_SCHEMA.scheme)
    identifier[
        ACTIONS_IDENTITY_SCHEMA.scheme["protocol"]
    ] = ACTIONS_IDENTITY_SCHEMA.protocol
    identifier[ACTIONS_IDENTITY_SCHEMA.scheme["identity"]] = str(identity.public_key)
    identifier[ACTIONS_IDENTITY_SCHEMA.scheme["reference"]] = str(reference)
    identifier[ACTIONS_IDENTITY_SCHEMA.scheme["signature"]] = base58.b58encode(
        signature
    ).decode()

    return ACTIONS_IDENTITY_SCHEMA.separator.join(identifier)


def validate_action_identifier_memo(
    identity: Pubkey, memos: Union[str, List[str], None]
) -> Union[bool, dict]:
    if not memos:
        return False

    if isinstance(memos, str):
        memos = memos.split(";")

    for memo in memos:
        try:
            memo = memo.strip()
            if memo.startswith("[") and "] " in memo:
                memo = memo.split("] ", 1)[1].strip()

            if not memo.count(":") >= 2:
                raise ActionIdentifierError("invalid memo formatting")

            identifier = memo.split(ACTIONS_IDENTITY_SCHEMA.separator)
            if len(identifier) != len(ACTIONS_IDENTITY_SCHEMA.scheme):
                raise ActionIdentifierError("invalid memo length")

            try:
                memo_identity = Pubkey(
                    identifier[ACTIONS_IDENTITY_SCHEMA.scheme["identity"]]
                )
            except ValueError:
                raise ActionIdentifierError("malformed memo identity")

            if not memo_identity:
                raise ActionIdentifierError("invalid memo identity")
            if str(memo_identity) != str(identity):
                raise ActionIdentifierError("identity mismatch")

            verified = nacl.signing.VerifyKey(bytes(identity)).verify(
                base58.b58decode(
                    identifier[ACTIONS_IDENTITY_SCHEMA.scheme["reference"]]
                ),
                base58.b58decode(
                    identifier[ACTIONS_IDENTITY_SCHEMA.scheme["signature"]]
                ),
            )

            if verified:
                return {
                    "verified": True,
                    "reference": identifier[
                        ACTIONS_IDENTITY_SCHEMA.scheme["reference"]
                    ],
                }
        except Exception:
            return False
    return False


async def verify_signature_info_for_identity(
    connection: AsyncClient, identity: Keypair, sig_info: dict
) -> bool:
    try:
        validated = validate_action_identifier_memo(
            identity.public_key, sig_info.get("memo")
        )
        if not validated:
            return False

        confirmed_sig_info = await find_reference(
            connection, Pubkey(validated["reference"])
        )

        if confirmed_sig_info["signature"] == sig_info["signature"]:
            return True
    except Exception:
        return False
    return False


def create_action_identifier_instruction(
    identity: Keypair, reference: Optional[Pubkey] = None
) -> dict:
    if reference is None:
        reference = Keypair().public_key

    memo = create_action_identifier_memo(identity, reference)

    return {
        "memo": memo,
        "reference": reference,
        "instruction": TransactionInstruction(
            program_id=Pubkey(MEMO_PROGRAM_ID), data=memo.encode("utf-8"), keys=[]
        ),
    }


def get_action_identity_from_env(env_key: str = "ACTION_IDENTITY_SECRET") -> Keypair:
    import json
    import os

    try:
        if env_key not in os.environ:
            raise ValueError("missing env key")
        return Keypair.from_secret_key(bytes(json.loads(os.environ[env_key])))
    except Exception:
        raise ValueError(f"invalid identity in env variable: '{env_key}'")
